<?php

namespace Drupal\webform_googlesheets\Plugin\WebformHandler;

use Google_Service_Sheets;
use Google_Service_Sheets_AppendCellsRequest;
use Google_Service_Sheets_BatchUpdateSpreadsheetRequest;
use Google_Service_Sheets_CellData;
use Google_Service_Sheets_ExtendedValue;
use Google_Service_Sheets_Request;
use Google_Service_Sheets_RowData;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\webform\Element\WebformManagedFileBase;
use Drupal\webform\Plugin\WebformElementManagerInterface;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\WebformSubmissionConditionsValidatorInterface;
use Drupal\webform\WebformSubmissionInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form submission to Google Sheets handler.
 *
 * @WebformHandler(
 *   id = "googlesheets",
 *   label = @Translation("Google Sheets"),
 *   category = @Translation("Google Sheets"),
 *   description = @Translation("Adds a form submission to a Google Sheets."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_PROCESSED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_OPTIONAL,
 *   tokens = TRUE,
 * )
 */
class WebformGoogleSheetsHanlder extends WebformHandlerBase {

  /**
   * A webform element plugin manager.
   *
   * @var \Drupal\webform\Plugin\WebformElementManagerInterface
   */
  protected $elementManager;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    LoggerChannelFactoryInterface $logger_factory,
    ConfigFactoryInterface $config_factory,
    EntityTypeManagerInterface $entity_type_manager,
    WebformSubmissionConditionsValidatorInterface $conditions_validator,
    WebformElementManagerInterface $element_manager
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $logger_factory,
      $config_factory,
      $entity_type_manager,
      $conditions_validator
    );
    $this->elementManager = $element_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('logger.factory'),
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('webform_submission.conditions_validator'),
      $container->get('plugin.manager.webform.element')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary() {
    $configuration = $this->getConfiguration();

    return [
      '#settings' => $configuration['settings'],
    ] + parent::getSummary();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $field_names = array_keys(
      \Drupal::service('entity_field.manager')
        ->getBaseFieldDefinitions('webform_submission')
    );

    return [
      'spreadsheet_url' => '',
      'spreadsheet_id' => '',
      'spreadsheet_sheet_id' => '',
      'excluded_fields' => array_combine($field_names, $field_names),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    parent::buildConfigurationForm($form, $form_state);

    // Google Sheets configuration.
    $form['googlesheets'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Google Sheets settings'),
    ];
    $form['googlesheets']['spreadsheet_url'] = [
      '#type' => 'url',
      '#title' => $this->t('Sheet URL'),
      '#required' => TRUE,
      '#default_value' => $this->configuration['spreadsheet_url'],
      '#empty_value' => '',
    ];

    // Submission data.
    $form['submission_data'] = [
      '#type' => 'details',
      '#title' => $this->t('Submission data'),
    ];
    // Display warning about file uploads.
    if ($this->getWebform()->hasManagedFile()) {
      $form['submission_data']['managed_file_message'] = [
        '#type' => 'webform_message',
        '#message_message' => $this->t('Upload files will include the file\'s id, name, uri, and data (<a href=":href">Base64</a> encode).', [':href' => 'https://en.wikipedia.org/wiki/Base64']),
        '#message_type' => 'warning',
        '#message_close' => TRUE,
        '#message_id' => 'webform_node.references',
        '#message_storage' => WebformMessage::STORAGE_SESSION,
      ];
    }
    $form['submission_data']['excluded_fields'] = [
      '#type' => 'webform_excluded_columns',
      '#title' => $this->t('Submitted data'),
      '#title_display' => 'invisible',
      '#webform_id' => $this->getWebform()->id(),
      '#required' => TRUE,
      '#default_value' => $this->configuration['excluded_fields'],
    ];

    return $this->setSettingsParents($form);
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::validateConfigurationForm($form, $form_state);

    if (
      !preg_match('#https?://docs.google.com/spreadsheets/d/(.+)/edit(\#gid=(\d+))?#', $form_state->getValue('spreadsheet_url'), $matches)
    ) {
      $form_state->setErrorByName('googlesheets][spreadsheet_url', $this->t('Please provide a valid Google Sheet URL.'));
    }
    else {
      $form_state->setValue('spreadsheet_id', $matches[1]);
      $form_state->setValue('spreadsheet_sheet_id', !empty($matches[3]) ? $matches[3] : 0);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    $this->applyFormStateToConfiguration($form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(WebformSubmissionInterface $webform_submission, $update = TRUE) {
    // If update operation, do nothing.
    if ($update) {
      return;
    }

    // Prepare request.
    $fields = $this->prepareFields($webform_submission);
    $rows = $this->prepareRow($fields);
    $request = $this->prepareRequest($this->configuration['spreadsheet_sheet_id'], $rows);

    // Execute request.
    $success = FALSE;
    try {
      $success = $this
        ->makeRequest($this->configuration['spreadsheet_id'], $request)
        ->valid();
    }
    catch (\Exception $e) {
      $this->getLogger('webform_googlesheets')->error($e->getMessage());
    }

    // Log results.
    if ($success) {
      $this->log($webform_submission, 'google sheets append success', $this->t('Webform submission had been successfully appended to Google Sheets.'));
    }
    else {
      $this->submissionStorage->log($webform_submission, [
        'handler_id' => $this->getHandlerId(),
        'operation' => 'google sheets append fail',
        'message' => $this->t('An error occurred appending webform submission to Google Sheets.'),
      ]);
    }
  }

  /**
   * Helper method to prepare webform submission fields.
   *
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   A webform submission.
   *
   * @return array
   *   An array with fields.
   */
  private function prepareFields(WebformSubmissionInterface $webform_submission) {
    // Get field information.
    $fields = $webform_submission->toArray(TRUE);

    // Flatten data and prioritize the element data over the
    // webform submission data.
    $fields_data = $fields['data'];
    unset($fields['data']);
    $fields = $fields_data + $fields;

    // Excluded selected submission fields.
    $fields = array_diff_key($fields, $this->configuration['excluded_fields']);

    // Append uploaded file name, uri, and base64 data to fields.
    $webform = $this->getWebform();
    foreach ($fields as $key => $value) {
      if (empty($value)) {
        continue;
      }

      $element = $webform->getElement($key);
      if (!$element) {
        continue;
      }

      $element_plugin = $this->elementManager->getElementInstance($element);
      if (!($element_plugin instanceof WebformManagedFileBase)) {
        continue;
      }

      if ($element_plugin->hasMultipleValues($element)) {
        foreach ($value as $fid) {
          $fields['_' . $key][] = $this->getResponseFileData($fid);
        }
      }
      else {
        $fields['_' . $key] = $this->getResponseFileData($value);
        // @deprecated in Webform 8.x-5.0-rc17. Use new format
        // The code needs to be removed before 8.x-5.0 or 8.x-6.x.
        $fields += $this->getResponseFileData($value, $key . '__');
      }
    }

    // Replace tokens.
    $fields = $this->replaceTokens($fields, $webform_submission);

    // Return values only.
    return array_values($fields);
  }

  /**
   * Helper method to prepare cell object.
   *
   * @param string $data
   *   A string to insert into a cell.
   *
   * @return \Google_Service_Sheets_CellData
   *   Cells to append in the row.
   */
  private function prepareCell($data) {
    $value = new Google_Service_Sheets_ExtendedValue();

    switch (gettype($data)) {
      case 'boolean':
        $value->setBoolValue($data);
        break;

      case 'integer':
      case 'double':
        $value->setNumberValue($data);
        break;

      case 'string':
        $value->setStringValue($data);
        break;

      default:
        // Set nothing?
        break;
    }

    $cell = new Google_Service_Sheets_CellData();
    $cell->setUserEnteredValue($value);

    return $cell;
  }

  /**
   * Helper method to prepare row object.
   *
   * @param array $data
   *   An array with row values.
   *
   * @return \Google_Service_Sheets_RowData
   *   Rows to append in the request.
   */
  private function prepareRow(array $data) {
    $row = new Google_Service_Sheets_RowData();
    $row->setValues(array_map([$this, 'prepareCell'], array_values($data)));

    return $row;
  }

  /**
   * Helper method to prepare Google API request.
   *
   * @param string $sheet_id
   *   A string with Google Spreadsheet Sheet ID.
   * @param \Google_Service_Sheets_RowData $rows
   *   Rows to append in the request.
   *
   * @return \Google_Service_Sheets_BatchUpdateSpreadsheetRequest
   *   Batch update request for a spreadsheet.
   */
  private function prepareRequest($sheet_id, Google_Service_Sheets_RowData $rows) {
    $append_request = new Google_Service_Sheets_AppendCellsRequest();
    $append_request->setSheetId($sheet_id);
    $append_request->setRows($rows);
    $append_request->setFields('userEnteredValue');

    $request = new Google_Service_Sheets_Request();
    $request->setAppendCells($append_request);

    return new Google_Service_Sheets_BatchUpdateSpreadsheetRequest(['requests' => [$request]]);
  }

  /**
   * Helper method to perform Google API request.
   *
   * @param string $spreadsheet_id
   *   A string with Google Spreadsheet ID.
   * @param \Google_Service_Sheets_BatchUpdateSpreadsheetRequest $request
   *   Batch update request for a spreadsheet.
   *
   * @return \Google_Service_Sheets_BatchUpdateSpreadsheetResponse
   *   Google API response.
   */
  private function makeRequest($spreadsheet_id, Google_Service_Sheets_BatchUpdateSpreadsheetRequest $request) {
    $service = new Google_Service_Sheets(
      \Drupal::service('google_api_client.client')->googleClient
    );

    return $service->spreadsheets->batchUpdate($spreadsheet_id, $request);
  }

}
